#include <fstream>

#include "ImgIO.h"

using namespace std;

GrayscaleImage readImage(const std::string& filename) {
    ifstream inFile(filename, ios_base::binary);
    if(!inFile.is_open()) {
        cout << "Check that the file is in your working directory!" << endl;
        throw logic_error("Error opening file");
    }
    //See https://en.wikipedia.org/wiki/BMP_file_format for details

    //magic bytes for bmp image to start with
    uint8_t header[] = {0x42, 0x4D};
    inFile.read(reinterpret_cast<char*>(header), 2);

    //byte 10 has the location of image data
    inFile.seekg(10);
    uint32_t data_start;
    inFile.read(reinterpret_cast<char*>(&data_start), 4);

    //bytes 18 and 22 are width and height
    inFile.seekg(18);
    uint32_t img_width, img_height;
    inFile.read(reinterpret_cast<char*>(&img_width), 4);
    inFile.read(reinterpret_cast<char*>(&img_height), 4);

    GrayscaleImage im(img_width, img_height);

    int bytesPerRow = (im.getWidth() * 3);
    int bytesInFinalWordOfRow = bytesPerRow % 4;
    int rowPaddingBytes = bytesInFinalWordOfRow == 0 ? 0 : 4 - bytesInFinalWordOfRow;
    int bytesPerRowPadded = bytesPerRow + rowPaddingBytes;

    for(int row = 0; row < im.getHeight(); row++) {
        //jump to start of row
        uint32_t rowStart = data_start + (im.getHeight() - 1 - row) * bytesPerRowPadded;
        inFile.seekg(rowStart);
        for(int col = 0; col < im.getWidth(); col++) {
            //real image is stored as RGB. We will read all of those and average them.
            uint8_t rgbVals[3];
            inFile.read(reinterpret_cast<char*>(rgbVals), 3);
            uint8_t grayVal = static_cast<uint8_t>( (rgbVals[0] + rgbVals[1] + rgbVals[2]) / 3 );
            im.setPixel(row, col, grayVal );
        }
    }
    if(inFile.fail()) {
        throw logic_error("Error reading file");
    }

    return im;
}

void writeImage(const GrayscaleImage& img, const std::string& filename) {
    //See https://en.wikipedia.org/wiki/BMP_file_format for details

    ofstream outFile(filename, ios_base::binary);

    int bytesPerRow = (img.getWidth() * 3);
    int bytesInFinalWordOfRow = bytesPerRow % 4;
    int rowPaddingBytes = bytesInFinalWordOfRow == 0 ? 0 : 4 - bytesInFinalWordOfRow;
    int bytesPerRowPadded = bytesPerRow + rowPaddingBytes;

    uint32_t totalBitmapBytes = bytesPerRowPadded * img.getHeight();

    uint32_t size = 14 + 40 + totalBitmapBytes;

    //magic bytes for start
    uint8_t header[] = {0x42, 0x4D};
    outFile.write(reinterpret_cast<char*>(header), 2);
    //file size
    outFile.write(reinterpret_cast<char*>(&size), 4);

    //padding
    uint16_t header2[] = {0, 0};
    outFile.write(reinterpret_cast<char*>(header2), 4);

    //data start
    uint32_t byteStart = 14 + 40;
    outFile.write(reinterpret_cast<char*>(&byteStart), 4);

    //header
    uint32_t dibSize = 40;
    uint32_t img_height = img.getHeight(), img_width = img.getWidth();
    outFile.write(reinterpret_cast<char*>(&dibSize), 4);
    outFile.write(reinterpret_cast<char*>(&img_width), 4);
    outFile.write(reinterpret_cast<char*>(&img_height), 4);

    uint16_t colorPlanes = 1;
    outFile.write(reinterpret_cast<char*>(&colorPlanes), 2);
    uint16_t bitsPerPixel = 24;
    outFile.write(reinterpret_cast<char*>(&bitsPerPixel), 2);
    uint32_t compression = 0;
    outFile.write(reinterpret_cast<char*>(&compression), 4);

    outFile.write(reinterpret_cast<char*>(&totalBitmapBytes), 4);

    uint32_t UNSPECIFIED = 0;   //unspecified
    outFile.write(reinterpret_cast<char*>(&UNSPECIFIED), 4);  //horizontal res
    outFile.write(reinterpret_cast<char*>(&UNSPECIFIED), 4);  //vertical res
    outFile.write(reinterpret_cast<char*>(&UNSPECIFIED), 4);  //colors in palate
    outFile.write(reinterpret_cast<char*>(&UNSPECIFIED), 4);  //important colors

    //data
    for(int i = img.getHeight() - 1; i >= 0; i--) {
       for(int j = 0; j < img.getWidth(); j++) {
           //write each pixel 3 times to make identical red/green/blue values
           uint8_t pixelValue = img.getPixel(i, j);
           outFile.write(reinterpret_cast<char*>(&pixelValue), 1);
           outFile.write(reinterpret_cast<char*>(&pixelValue), 1);
           outFile.write(reinterpret_cast<char*>(&pixelValue), 1);
       }
       //align to 4 bytes
       outFile.write(reinterpret_cast<char*>(&UNSPECIFIED), rowPaddingBytes);
    }
}
